@@ -12,10 +12,9 @@ from feeluown.utils import aio
 from feeluown.utils.reader import wrap
 from feeluown.media import Media, MediaType
 from feeluown.excs import ProviderIOError
-from feeluown.library import ModelState, NotSupported, ModelFlags
-from feeluown.models import reverse, ModelType
+from feeluown.library import ModelState, NotSupported, ModelFlags, ModelType
 
-from feeluown.gui.helpers import async_run, BgTransparentMixin, \
+from feeluown.gui.helpers import BgTransparentMixin, \
     disconnect_slots_if_has, fetch_cover_wrapper
 from feeluown.gui.widgets.menu import SongMenuInitializer
 from feeluown.gui.widgets.imglist import ImgListView
@@ -54,8 +53,6 @@ class Renderer:
         # pylint: disable=protected-access
         self._app = container._app
 
-        self.real_show_model = container.show_model
-
     async def render(self):
         """render contents in table container
 
@@ -92,9 +89,6 @@ class Renderer:
                 self.meta_widget.set_cover_pixmap(pixmap)
             self._app.ui.table_container.updateGeometry()
 
-    def show_model(self, model):
-        aio.create_task(self.real_show_model(model))
-
     def show_albums(self, reader):
         self._show_model_with_cover(reader,
                                     self.albums_table,
@@ -124,7 +118,7 @@ class Renderer:
         filter_model = filter_model_cls()
         source_name_map = {p.identifier: p.name for p in self._app.library.list()}
         model = model_cls(reader,
-                          fetch_cover_wrapper(self._app.img_mgr),
+                          fetch_cover_wrapper(self._app),
                           source_name_map=source_name_map)
         filter_model.setSourceModel(model)
         table.setModel(filter_model)
@@ -165,43 +159,6 @@ class Renderer:
         self.comments_table.setModel(model)
 
 
-class PlaylistRenderer(Renderer):
-    def __init__(self, playlist):
-        self.playlist = playlist
-
-    async def render(self):
-        playlist = self.playlist
-
-        # show playlist title
-        self.meta_widget.show()
-        self.meta_widget.title = playlist.name
-
-        # show playlist song list
-        with suppress(ProviderIOError):
-            if playlist.meta.allow_create_songs_g:
-                reader = wrap(playlist.create_songs_g())
-            else:
-                songs = await async_run(lambda: playlist.songs)
-                reader = wrap(songs)
-            self.show_songs(reader=reader, show_count=True)
-
-        # show playlist cover
-        if playlist.cover:
-            aio.create_task(
-                self.show_cover(playlist.cover, reverse(playlist, '/cover')))
-
-        def remove_song(song):
-            playlist.remove(song.identifier)
-
-        self.songs_table.remove_song_func = remove_song
-        self.tabbar.show_desc_needed.connect(lambda: aio.create_task(self._show_desc()))
-
-    async def _show_desc(self):
-        with suppress(ProviderIOError):
-            desc = await async_run(lambda: self.playlist.desc)
-            self.show_desc(desc)
-
-
 class SongsCollectionRenderer(Renderer):
     def __init__(self, collection):
         self.collection = collection
@@ -212,7 +169,7 @@ class SongsCollectionRenderer(Renderer):
         self.meta_widget.title = collection.name
         self.meta_widget.updated_at = collection.updated_at
         self.meta_widget.created_at = collection.created_at
-        self.songs_table.remove_song_func = collection.remove
+        self.songs_table.remove_song_func = self.remove_song
         self._show_songs()
 
         # only show tabbar if description is not empty
@@ -229,6 +186,11 @@ class SongsCollectionRenderer(Renderer):
         self.show_songs(wrap([model for model in self.collection.models
                               if model.meta.model_type == ModelType.song]))
 
+    def remove_song(self, song):
+        self.collection.remove(song)
+        # Re-render songs table so that user can see that the song is removed.
+        self._show_songs()
+
 
 class AlbumsCollectionRenderer(Renderer):
     def __init__(self, reader):
@@ -296,9 +258,9 @@ class TableContainer(QFrame, BgTransparentMixin):
         self._tables.append(self.comments_table)
 
         self.songs_table.play_song_needed.connect(
-            lambda song: asyncio.ensure_future(self.play_song(song)))
+            lambda song: aio.run_afn(self.play_song, song))
         self.videos_table.play_video_needed.connect(
-            lambda video: aio.create_task(self.play_video(video)))
+            self._app.playlist.play_model)
 
         def goto_model(model):
             self._app.browser.goto(model=model)
@@ -311,7 +273,8 @@ class TableContainer(QFrame, BgTransparentMixin):
                        ]:
             signal.connect(goto_model)
 
-        self.toolbar.play_all_needed.connect(self.play_all)
+        self.toolbar.play_all_needed.connect(
+            lambda: aio.run_afn(self.play_all))
         self.songs_table.add_to_playlist_needed.connect(self._add_songs_to_playlist)
         self.songs_table.about_to_show_menu.connect(self._songs_table_about_to_show_menu)
         self.songs_table.activated.connect(
@@ -426,41 +389,27 @@ class TableContainer(QFrame, BgTransparentMixin):
         await self._renderer.render()
 
     async def play_song(self, song):
-        self._app.player.play_song(song)
+        self._app.playlist.play_model(song)
 
-    async def play_video(self, video):
-        media = await aio.run_in_executor(None, lambda: video.media)
-        self._app.player.play(media)
-
-    def play_all(self):
+    async def play_all(self):
         task_name = 'play-all'
         task_spec = self._app.task_mgr.get_or_create(task_name)
-
-        def reader_readall_cb(task):
-            with suppress(ProviderIOError, asyncio.CancelledError):
-                songs = task.result()
-                self._app.player.play_songs(songs=songs)
-            self.toolbar.enter_state_playall_end()
-
         model = self.songs_table.model()
-        # FIXME: think about a more elegant way
+        # FIXME: think about a more elegant way to get reader.
         reader = model.sourceModel().reader
         if reader is not None and reader.count is not None:
-            task = task_spec.bind_blocking_io(reader.readall)
             self.toolbar.enter_state_playall_start()
-            task.add_done_callback(reader_readall_cb)
+            with suppress(ProviderIOError, asyncio.CancelledError):
+                songs = await task_spec.bind_blocking_io(reader.readall)
+                self._app.playlist.set_models(songs)
+                task = self._app.playlist.next()
+                if task is not None:
+                    await task
+                    self._app.player.resume()
+            self.toolbar.enter_state_playall_end()
             return
         assert False, 'The play_all_btn should be hide in this page'
 
-    async def show_model(self, model):
-        model = self._app.library.cast_model_to_v1(model)
-        model_type = ModelType(model.meta.model_type)
-        if model_type == ModelType.playlist:
-            renderer = PlaylistRenderer(model)
-        else:
-            renderer = None
-        await self.set_renderer(renderer)
-
     def show_collection(self, coll):
         renderer = SongsCollectionRenderer(coll)
         aio.create_task(self.set_renderer(renderer))
@@ -521,8 +470,8 @@ class TableContainer(QFrame, BgTransparentMixin):
                         reader = model.reader
                         songs = await aio.run_fn(reader.readall)
                         self._app.playlist.clear()
-                        self._app.playlist.init_from(songs)
-            self._app.player.play_song(song)
+                        self._app.playlist.set_models(songs)
+            self._app.playlist.play_model(song)
         else:
             try:
                 song = await aio.run_in_executor(
