@@ -8,12 +8,13 @@ from PyQt5.QtWidgets import (
     QPushButton, QSizePolicy, QMenu,
 )
 
+from feeluown.utils import aio
 from feeluown.excs import ProviderIOError
 from feeluown.media import MediaType
 from feeluown.player import PlaybackMode, State
 from feeluown.gui.widgets.lyric import Window as LyricWindow
 from feeluown.gui.widgets.menu import SongMenuInitializer
-from feeluown.gui.helpers import async_run, resize_font
+from feeluown.gui.helpers import resize_font
 from feeluown.gui.widgets import TextButton
 from feeluown.gui.widgets.playlist_button import PlaylistButton
 from feeluown.gui.widgets.volume_button import VolumeButton
@@ -310,20 +311,16 @@ class PlayerControlPanel(QFrame):
             lambda volume: setattr(self._app.player, 'volume', volume))
 
         player = self._app.player
-
-        player.state_changed.connect(self._on_player_state_changed,
-                                     aioqueue=True)
-        player.playlist.playback_mode_changed.connect(
-            self.on_playback_mode_changed, aioqueue=True)
-        player.playlist.song_changed.connect(
-            self.on_player_song_changed, aioqueue=True)
-        player.media_changed.connect(
-            self.on_player_media_changed, aioqueue=True)
+        playlist = self._app.playlist
+        playlist.playback_mode_changed.connect(self.on_playback_mode_changed,
+                                               aioqueue=True)
+        playlist.song_changed.connect(self.on_player_song_changed, aioqueue=True)
+        player.state_changed.connect(self._on_player_state_changed, aioqueue=True)
+        player.metadata_changed.connect(self.on_metadata_changed, aioqueue=True)
         player.volume_changed.connect(self.volume_btn.on_volume_changed)
-        self._app.live_lyric.sentence_changed.connect(
-            self.lyric_window.set_sentence)
-        self.lyric_window.play_previous_needed.connect(player.playlist.previous)
-        self.lyric_window.play_next_needed.connect(player.playlist.next)
+        self._app.live_lyric.sentence_changed.connect(self.lyric_window.set_sentence)
+        self.lyric_window.play_previous_needed.connect(playlist.previous)
+        self.lyric_window.play_next_needed.connect(playlist.next)
 
         self._update_pms_btn_text()
         self._setup_ui()
@@ -400,7 +397,7 @@ class PlayerControlPanel(QFrame):
         self._layout.setContentsMargins(0, 0, 0, 0)
 
     def _switch_playback_mode(self):
-        playlist = self._app.player.playlist
+        playlist = self._app.playlist
         pm_total = len(self._playback_modes)
         pm_idx = self._playback_modes.index(playlist.playback_mode)
         if pm_idx < pm_total - 1:
@@ -409,52 +406,69 @@ class PlayerControlPanel(QFrame):
             pm_idx = 0
         playlist.playback_mode = self._playback_modes[pm_idx]
 
-    def on_playback_mode_changed(self, playback_mode):
+    def on_playback_mode_changed(self, _):
         self._update_pms_btn_text()
 
     def _update_pms_btn_text(self):
-        playback_mode = self._app.player.playlist.playback_mode
+        playback_mode = self._app.playlist.playback_mode
         alias = self._pm_alias_map[playback_mode]
         self.pms_btn.setText(alias)
 
     def on_player_song_changed(self, song):
-        if song is None:
-            self.song_source_label.setText('歌曲来源')
-            self.song_title_label.setText('No song is playing.')
-            return
-        source_name_map = {p.identifier: p.name
-                           for p in self._app.library.list()}
-        text = '{} - {}'.format(song.title_display, song.artists_name_display)
-        # width -> three button + source label + text <= progress slider
-        # three button: 63, source label: 150
-        source_name = source_name_map.get(song.source, song.source)
-        self.song_source_label.setText(source_name)
-        self.song_title_label.setText(text)
         task_spec = self._app.task_mgr.get_or_create('update-mv-btn-status')
         task_spec.bind_coro(self.update_mv_btn_status(song))
 
-    def on_player_media_changed(self, media):
+    def on_metadata_changed(self, metadata):
+        if not metadata:
+            self.song_source_label.setText('歌曲来源')
+            self.song_title_label.setText('')
+            return
+
+        # Set main text.
+        text = metadata.get('title', '')
+        if text:
+            artists = metadata.get('artists', [])
+            if artists:
+                # FIXME: use _get_artists_name
+                text += f" - {','.join(artists)}"
+        self.song_title_label.setText(text)
+
+        # Set source name.
+        source = metadata.get('source', '')
+        default = '未知来源'
+
+        if source:
+            source_name_map = {p.identifier: p.name
+                               for p in self._app.library.list()}
+            name = source_name_map.get(source, default)
+            self.song_source_label.setText(name)
+        else:
+            self.song_source_label.setText(default)
+
+        # Set audio bitrate info if available.
+        media = self._app.player.current_media
         if media is not None and media.type_ == MediaType.audio:
-            metadata = media.metadata
-            if metadata.bitrate:
+            props = media.props
+            if props.bitrate:
                 text = self.song_source_label.text()
-                bitrate_text = str(metadata.bitrate) + 'kbps'
-                self.song_source_label.setText(
-                    '{} - {}'.format(text, bitrate_text))
+                self.song_source_label.setText(f'{text} - {props.bitrate}kbps')
 
     async def update_mv_btn_status(self, song):
-        song = self._app.library.cast_model_to_v1(song)
+        if song is None:
+            self.mv_btn.setEnabled(False)
+            return
+
         try:
-            mv = await async_run(lambda: song.mv)
+            mv = await aio.run_fn(self._app.library.song_get_mv, song)
         except ProviderIOError:
             logger.exception('fetch song mv info failed')
             self.mv_btn.setEnabled(False)
         else:
-            if mv:
-                self.mv_btn.setToolTip(mv.name)
-                self.mv_btn.setEnabled(True)
-            else:
+            if mv is None:
                 self.mv_btn.setEnabled(False)
+            else:
+                self.mv_btn.setToolTip(mv.title)
+                self.mv_btn.setEnabled(True)
 
     def _on_player_state_changed(self, state):
         self.pp_btn.setChecked(state == State.playing)
