@@ -1,30 +1,32 @@
 import logging
 from functools import partial, lru_cache
-from typing import List, cast
+from typing import List, cast, Optional, Union
 
-from feeluown.utils import aio
-from feeluown.utils.dispatch import Signal
 from feeluown.media import Media
 from feeluown.models import SearchType, ModelType
+from feeluown.utils import aio
+from feeluown.utils.dispatch import Signal
 from feeluown.utils.utils import log_exectime
+from feeluown.utils.reader import create_reader
 from .provider import AbstractProvider
 from .provider_v2 import ProviderV2
 from .excs import (
     NotSupported, MediaNotFound, NoUserLoggedIn, ProviderAlreadyExists,
-    ProviderNotFound,
+    ProviderNotFound, ModelNotFound
 )
 from .flags import Flags as PF
 from .models import (
-    ModelFlags as MF, BaseModel, BriefSongModel, SongModel, UserModel,
+    ModelFlags as MF, BaseModel, BriefSongModel, UserModel,
+    get_modelcls_by_type, V2SupportedModelTypes,
 )
 from .model_protocol import (
-    ModelProtocol, BriefSongProtocol, SongProtocol, UserProtocol,
+    BriefVideoProtocol, ModelProtocol, BriefSongProtocol, SongProtocol, UserProtocol,
+    LyricProtocol, VideoProtocol, BriefAlbumProtocol, BriefArtistProtocol
 )
 
 
 logger = logging.getLogger(__name__)
 
-
 FULL_SCORE = 10
 MIN_SCORE = 5
 
@@ -106,6 +108,14 @@ def _get_display_property_or_raise(model, attr):
     return getattr(model, f'_display_store_{attr}')
 
 
+def err_provider_not_support_flag(pid, model_type, op):
+    op_str = str(op)
+    if op is PF.get:
+        op_str = 'get'
+    mtype_str = str(ModelType(model_type))
+    return NotSupported(f"provider:{pid} does't support '{op_str}' for {mtype_str}")
+
+
 class Library:
     """音乐库，管理资源提供方以及资源"""
 
@@ -132,7 +142,7 @@ class Library:
         >>> library.register(dummy_provider)
         Traceback (most recent call last):
             ...
-        feeluown.library.excs.ProviderAlreadyExists
+        feeluown.excs.ProviderAlreadyRegistered
         """
         if not isinstance(provider, AbstractProvider):
             raise ValueError('invalid provider instance')
@@ -151,7 +161,7 @@ class Library:
         else:
             self.provider_removed.emit(provider)
 
-    def get(self, identifier):
+    def get(self, identifier) -> Optional[Union[AbstractProvider, ProviderV2]]:
         """通过资源提供方唯一标识获取提供方实例"""
         for provider in self._providers:
             if provider.identifier == identifier:
@@ -213,7 +223,9 @@ class Library:
                 logger.exception('search task failed')
                 continue
             else:
-                yield result
+                # When a provider does not implement search method, it returns None.
+                if result is not None:
+                    yield result
 
     @log_exectime
     def list_song_standby(self, song, onlyone=True):
@@ -345,7 +357,7 @@ class Library:
 
     # provider common
 
-    def get_or_raise(self, identifier) -> ProviderV2:
+    def get_or_raise(self, identifier) -> Union[AbstractProvider, ProviderV2]:
         """
         :raises ProviderNotFound:
         """
@@ -354,6 +366,13 @@ class Library:
             raise ProviderNotFound(f'provider {identifier} not found')
         return provider
 
+    def getv2_or_raise(self, identifier) -> ProviderV2:
+        provider = self.get_or_raise(identifier)
+        # You should ensure the provider is v2 first. For example, if check_flags
+        # returns true, the provider must be a v2 instance.
+        assert isinstance(provider, ProviderV2), 'provider must be v2'
+        return provider
+
     def check_flags(self, source: str, model_type: ModelType, flags: PF) -> bool:
         """Check if a provider satisfies the specific ability for a model type
 
@@ -369,6 +388,7 @@ class Library:
             return False
         if isinstance(provider, ProviderV2):
             return provider.check_flags(model_type, flags)
+        # Always return false when the provider is not a v2 instance.
         return False
 
     def check_flags_by_model(self, model: ModelProtocol, flags: PF) -> bool:
@@ -385,6 +405,8 @@ class Library:
 
         During the model migration from v1 to v2, v2 may lack some ability.
         Cast the model to v1 to acquire such ability.
+
+        :raises NotSupported: provider doesn't support v1 model
         """
         if isinstance(model, BaseModel) and (model.meta.flags & MF.v2):
             return self._cast_model_to_v1_impl(model)
@@ -394,6 +416,11 @@ class Library:
     def _cast_model_to_v1_impl(self, model):
         provider = self.get_or_raise(model.source)
         ModelCls = provider.get_model_cls(model.meta.model_type)
+        # The source of the default SongModel is None. When ModelCls.source
+        # is None, it means that the provider does not has its own model class.
+        if ModelCls.source is None:
+            model_type_str = str(ModelType(model.meta.model_type))
+            raise NotSupported(f'provider has no v1 model impl for {model_type_str}')
         kv = {}
         for field in ModelCls.meta.fields_display:
             kv[field] = getattr(model, field)
@@ -403,38 +430,25 @@ class Library:
     # Songs
     # -----
     def song_upgrade(self, song: BriefSongProtocol) -> SongProtocol:
-        if song.meta.flags & MF.v2:
-            if MF.normal in song.meta.flags:
-                upgraded_song = cast(SongProtocol, song)
-            else:
-                provider = self.get_or_raise(song.source)
-                if self.check_flags_by_model(song, PF.get):
-                    upgraded_song = provider.song_get(song.identifier)
-                else:
-                    raise NotSupported("provider has not flag 'get' for 'song'")
-        else:
-            fields = [f for f in list(SongModel.__fields__)
-                      if f not in list(BaseModel.__fields__)]
-            for field in fields:
-                getattr(song, field)
-            upgraded_song = cast(SongProtocol, song)
-        return upgraded_song
+        return self._model_upgrade(song)  # type: ignore
 
     def song_list_similar(self, song: BriefSongProtocol) -> List[BriefSongProtocol]:
-        provider = self.get_or_raise(song.source)
+        provider = self.getv2_or_raise(song.source)
         return provider.song_list_similar(song)
 
     def song_list_hot_comments(self, song: BriefSongProtocol):
-        provider = self.get_or_raise(song.source)
+        provider = self.getv2_or_raise(song.source)
         return provider.song_list_hot_comments(song)
 
     def song_prepare_media(self, song: BriefSongProtocol, policy) -> Media:
         provider = self.get(song.source)
         if provider is None:
+            # FIXME: raise ProviderNotfound
             raise MediaNotFound(f'provider:{song.source} not found')
         if song.meta.flags & MF.v2:
             # provider MUST has multi_quality flag for song
             assert self.check_flags_by_model(song, PF.multi_quality)
+            assert isinstance(provider, ProviderV2)
             media, _ = provider.song_select_media(song, policy)
         else:
             if song.meta.support_multi_quality:
@@ -454,9 +468,6 @@ class Library:
 
         .. versionadded:: 3.7.5
         """
-        provider = self.get(song.source)
-        if provider is None:
-            raise MediaNotFound(f'provider:{song.source} not found')
         song_v1 = self.cast_model_to_v1(song)
         mv = song_v1.mv
         if mv.meta.support_multi_quality:
@@ -471,16 +482,256 @@ class Library:
             raise MediaNotFound
         return media
 
-    def song_get_lyric(self, song: BriefSongModel):
-        pass
+    def song_get_mv(self, song: BriefSongProtocol) -> Optional[VideoProtocol]:
+        """
 
-    def song_get_mv(self, song: BriefSongModel):
-        pass
+        :raises NotSupported:
+        :raises ProviderNotFound:
+        """
+        provider = self.get_or_raise(song.source)
+        if self.check_flags_by_model(song, PF.mv):
+            assert isinstance(provider, ProviderV2)
+            mv = provider.song_get_mv(song)
+        else:
+            song_v1 = self.cast_model_to_v1(song)
+            mv = song_v1.mv
+            mv = cast(Optional[VideoProtocol], mv)
+        return mv
+
+    def song_get_lyric(self, song: BriefSongModel) -> Optional[LyricProtocol]:
+        """
+
+        Return None when lyric does not exist instead of raising exceptions,
+        because it is predictable.
+
+        :raises NotSupported:
+        :raises ProviderNotFound:
+        """
+        provider = self.get_or_raise(song.source)
+        if self.check_flags_by_model(song, PF.lyric):
+            assert isinstance(provider, ProviderV2)
+            lyric = provider.song_get_lyric(song)
+        else:
+            song_v1 = self.cast_model_to_v1(song)
+            lyric_v1 = song_v1.lyric
+            lyric = cast(Optional[LyricProtocol], lyric_v1)
+        return lyric
 
     def song_get_web_url(self, song: BriefSongProtocol) -> str:
-        provider = self.get_or_raise(song.source)
+        provider = self.getv2_or_raise(song.source)
         return provider.song_get_web_url(song)
 
+    # --------
+    # Album
+    # --------
+    def album_upgrade(self, album: BriefAlbumProtocol):
+        return self._model_upgrade(album)
+
+    # --------
+    # Artist
+    # --------
+    def artist_upgrade(self, artist: BriefArtistProtocol):
+        return self._model_upgrade(artist)
+
+    def artist_create_songs_rd(self, artist):
+        """Create songs reader for artist model.
+
+        :raises NotSupported:
+        """
+        provider = self.get_or_raise(artist.source)
+        if self.check_flags_by_model(artist, PF.songs_rd):
+            assert isinstance(provider, ProviderV2)
+            reader = provider.artist_create_songs_rd(artist)
+        else:
+            artist = self.cast_model_to_v1(artist)
+            if artist.meta.allow_create_songs_g:
+                reader = create_reader(artist.create_songs_g())
+            else:
+                reader = create_reader(artist.songs)
+        return reader
+
+    def artist_create_albums_rd(self, artist):
+        """Create albums reader for artist model.
+        """
+        provider = self.get_or_raise(artist.source)
+        if self.check_flags_by_model(artist, PF.albums_rd):
+            assert isinstance(provider, ProviderV2)
+            reader = provider.artist_create_albums_rd(artist)
+        else:
+            artist = self.cast_model_to_v1(artist)
+            if artist.meta.allow_create_albums_g:
+                reader = create_reader(artist.create_albums_g())
+            else:
+                raise NotSupported("can't create albums reader for artist")
+        return reader
+
+    # --------
+    # Playlist
+    # --------
+
+    def playlist_upgrade(self, playlist):
+        return self._model_upgrade(playlist)
+
+    def playlist_create_songs_rd(self, playlist):
+        """Create songs reader for artist model.
+
+        :raises NotSupported:
+        """
+        provider = self.get_or_raise(playlist.source)
+        if self.check_flags_by_model(playlist, PF.songs_rd):
+            assert isinstance(provider, ProviderV2)
+            reader = provider.playlist_create_songs_rd(playlist)
+        else:
+            playlist = self.cast_model_to_v1(playlist)
+            if playlist.meta.allow_create_songs_g:
+                reader = create_reader(playlist.create_songs_g())
+            else:
+                reader = create_reader(playlist.songs)
+        return reader
+
+    def playlist_remove_song(self, playlist, song) -> bool:
+        """Remove a song from the playlist
+
+        :return: true if the song is not in playlist anymore.
+        """
+        provider = self.get_or_raise(playlist.source)
+        if self.check_flags_by_model(playlist, PF.remove_song):
+            assert isinstance(provider, ProviderV2)
+            ok = provider.playlist_remove_song(playlist, song)
+        else:
+            playlist = self.cast_model_to_v1(playlist)
+            ok = playlist.remove(song.identifier)
+        return ok
+
+    def playlist_add_song(self, playlist, song) -> bool:
+        """Add a song to the playlist
+
+        :return: true if the song exists in playlist.
+        """
+        provider = self.get_or_raise(playlist.source)
+        if self.check_flags_by_model(playlist, PF.remove_song):
+            assert isinstance(provider, ProviderV2)
+            ok = provider.playlist_add_song(playlist, song)
+        else:
+            playlist = self.cast_model_to_v1(playlist)
+            ok = playlist.add(song.identifier)
+        return ok
+
+    # -------------------------
+    # generic methods for model
+    # -------------------------
+    def model_get(self, pid, mtype, mid):
+        """Get a (normal) model instance.
+
+        :param pid: provider id
+        :param mtype: model type
+        :param mid: model id
+        :return: model
+
+        :raise NotSupported: provider has not .get for this model type
+        :raise ResourceNotFound: model does not exist
+        """
+        provider = self.get_or_raise(pid)
+        model = None
+        try_v1way = True
+        if isinstance(provider, ProviderV2):
+            if provider.check_flags(mtype, PF.model_v2):
+                if provider.check_flags(mtype, PF.get):
+                    try_v1way = False
+                    model = provider.model_get(mtype, mid)
+
+        # Try to use the ModelV1.get API to get the model.
+        if try_v1way and isinstance(provider, AbstractProvider):
+            try:
+                model_cls = provider.get_model_cls(mtype)
+                model = model_cls.get(mid)
+            except AttributeError:
+                pass
+        if model is None:
+            raise ModelNotFound
+        return model
+
+    def model_get_cover(self, model):
+        """Get the cover url of model
+
+        :param model: model which has a 'cover' field.
+        :return: cover url if exists, else ''.
+        """
+        cover = ''
+        if MF.v2 in model.meta.flags:
+            if MF.normal in model.meta.flags:
+                cover = model.cover
+            else:
+                # TODO: upgrade artist model.
+                # Currently supported model types: (ModelType.album, ModelType.video).
+                if ModelType(model.meta.model_type) in V2SupportedModelTypes:
+                    um = self._model_upgrade(model)
+                    # FIXME: remove this hack lator.
+                    if ModelType(model.meta.model_type) is ModelType.artist:
+                        cover = um.pic_url
+                    else:
+                        cover = um.cover
+        else:
+            cover = model.cover
+            # Check if cover is a media object.
+            if cover and not isinstance(cover, str):
+                cover = cover.url
+        return cover
+
+    def _model_upgrade(self, model):
+        """
+        Thinking: currently, the caller must catch the NotSupported error.
+        """
+        try_v1way = True
+        upgraded_model = None
+        if model.meta.flags & MF.v2:
+            if MF.normal in model.meta.flags:
+                upgraded_model = model
+                try_v1way = False
+            else:
+                provider = self.getv2_or_raise(model.source)
+                # When the provider does not support 'get' for this model.
+                # Do not raise NotSupported here and try to use the v1 way.
+                #
+                # For example, provider X may support 'get' for SongModel, then
+                # the song.artists can return list of BriefArtistModel.
+                if self.check_flags_by_model(model, PF.get):
+                    upgraded_model = provider.model_get(
+                        model.meta.model_type, model.identifier)
+                    try_v1way = False
+
+        if try_v1way is True:
+            v1model = self.cast_model_to_v1(model)
+            modelcls = get_modelcls_by_type(ModelType(model.meta.model_type))
+            fields = [f for f in list(modelcls.__fields__)
+                      if f not in list(BaseModel.__fields__)]
+            for field in fields:
+                getattr(v1model, field)
+            upgraded_model = v1model
+        else:
+            assert upgraded_model is not None
+        return upgraded_model
+
+    # --------
+    # Video
+    # --------
+    def video_prepare_media(self, video: BriefVideoProtocol, policy) -> Media:
+        provider = self.get_or_raise(video.source)
+        if video.meta.flags & MF.v2:
+            # provider MUST has multi_quality flag for video
+            assert self.check_flags_by_model(video, PF.multi_quality)
+            assert isinstance(provider, ProviderV2)
+            media, _ = provider.video_select_media(video, policy)
+        else:
+            # V1 VideoModel has attribute `media`
+            if video.meta.support_multi_quality:
+                media, _ = video.select_media(policy)  # type: ignore
+            else:
+                media = video.media  # type: ignore
+        if not media:
+            raise MediaNotFound
+        return media
+
     # --------
     # Provider
     # --------
@@ -493,6 +744,7 @@ class Library:
         """
         provider = self.get_or_raise(source)
         if self.check_flags(source, ModelType.none, PF.current_user):
+            assert isinstance(provider, ProviderV2)
             return provider.has_current_user()
 
         try:
@@ -514,6 +766,7 @@ class Library:
         """
         provider = self.get_or_raise(source)
         if self.check_flags(source, ModelType.none, PF.current_user):
+            assert isinstance(provider, ProviderV2)
             return provider.get_current_user()
 
         user_v1 = getattr(provider, '_user', None)
