@@ -1,9 +1,11 @@
+from .metadata import MetadataFields, Metadata
 from .playlist import PlaybackMode
 from .base_player import State
 from .mpvplayer import MpvPlayer as Player
 from .playlist import PlaylistMode, Playlist
 from .fm import FM
 from .radio import SongRadio
+from .lyric import LiveLyric, parse_lyric_text
 
 
 __all__ = (
@@ -16,4 +18,10 @@ __all__ = (
 
     'Player',
     'Playlist',
+
+    'Metadata',
+    'MetadataFields',
+
+    'LiveLyric',
+    'parse_lyric_text',
 )
