@@ -2,6 +2,7 @@ import locale
 import logging
 
 from feeluown.utils.utils import use_mpv_old
+from .metadata import MetadataFields, Metadata
 
 if use_mpv_old():
     from mpv_old import (  # type: ignore
@@ -23,7 +24,7 @@ else:
     )
 
 from feeluown.utils.dispatch import Signal
-from feeluown.media import Media
+from feeluown.media import Media, VideoAudioManifest
 from .base_player import AbstractPlayer, State
 
 
@@ -38,8 +39,11 @@ class MpvPlayer(AbstractPlayer):
 
     todo: make me singleton
     """
-    def __init__(self, audio_device=b'auto', winid=None, *args, **kwargs):
-        super(MpvPlayer, self).__init__(*args, **kwargs)
+    def __init__(self, _=None, audio_device=b'auto', winid=None, **kwargs):
+        """
+        :param _: keep this arg to keep backward compatibility
+        """
+        super().__init__(**kwargs)
         # https://github.com/cosven/FeelUOwn/issues/246
         locale.setlocale(locale.LC_NUMERIC, 'C')
         mpvkwargs = {}
@@ -88,24 +92,63 @@ class MpvPlayer(AbstractPlayer):
         logger.debug('Player initialize finished.')
 
     def shutdown(self):
-        self._mpv.terminate()
+        # The mpv has already been terminated.
+        # The mpv can't terminate twice.
+        if self._mpv.handle is not None:
+            self._mpv.terminate()
 
-    def play(self, media, video=True):
-        logger.debug("Player will play: '%s'", media)
-        if isinstance(media, Media):
-            media = media
-        else:  # media is a url
-            media = Media(media)
-        self._set_http_headers(media.http_headers)
-        url = media.url
+    def play(self, media, video=True, metadata=None):
+        if video is False:
+            _mpv_set_property_string(self._mpv.handle, b'vid', b'no')
+        else:
+            _mpv_set_property_string(self._mpv.handle, b'vid', b'auto')
 
-        # Clear playlist before play next song,
-        # otherwise, mpv will seek to the last position and play.
         self.media_about_to_changed.emit(self._current_media, media)
-        self._mpv.playlist_clear()
-        self._mpv.play(url)
+        if media is None:
+            self._stop_mpv()
+        else:
+            logger.debug("Player will play: '%s'", media)
+            if isinstance(media, Media):
+                media = media
+            else:  # media is a url(string)
+                media = Media(media)
+            self._set_http_headers(media.http_headers)
+            self._stop_mpv()
+            if media.manifest is None:
+                url = media.url
+                # Clear playlist before play next song,
+                # otherwise, mpv will seek to the last position and play.
+                self._mpv.play(url)
+            elif isinstance(media.manifest, VideoAudioManifest):
+                video_url = media.manifest.video_url
+                audio_url = media.manifest.audio_url
+
+                def add_audio():
+                    try:
+                        if self.current_media is media:
+                            self._mpv.audio_add(audio_url)
+                            self.resume()
+                    finally:
+                        self.media_loaded.disconnect(add_audio)
+
+                if video is True:
+                    self._mpv.play(video_url)
+                    # It seems we can only add audio after the video is loaded.
+                    # TODO: add method connect_once for signal
+                    self.media_loaded.connect(add_audio, weak=False)
+                else:
+                    self._mpv.play(audio_url)
+            else:
+                assert False, 'Unknown manifest'
         self._current_media = media
         self.media_changed.emit(media)
+        if metadata is None:
+            self._current_metadata = {}
+        else:
+            # The metadata may be set by manual or automatic
+            metadata['__setby__'] = 'manual'
+            self._current_metadata = metadata
+        self.metadata_changed.emit(self.current_metadata)
 
     def set_play_range(self, start=None, end=None):
         if self._version >= (1, 28):
@@ -135,9 +178,8 @@ class MpvPlayer(AbstractPlayer):
     def stop(self):
         self._mpv.pause = True
         self.state = State.stopped
-        self._current_media = None
-        self._mpv.playlist_clear()
-        logger.info('Player stopped.')
+        self.play(None)
+        logger.debug('Player stopped.')
 
     @property
     def position(self):
@@ -148,6 +190,7 @@ class MpvPlayer(AbstractPlayer):
         if self._current_media:
             self._mpv.seek(position, reference='absolute')
             self._position = position
+            self.seeked.emit(position)
         else:
             logger.warn("can't set position when current media is empty")
 
@@ -165,6 +208,12 @@ class MpvPlayer(AbstractPlayer):
         self._video_format = vformat
         self.video_format_changed.emit(vformat)
 
+    def _stop_mpv(self):
+        # Remove current media.
+        self._mpv.play("")
+        # Clear the playlist so that no other media will be played.
+        self._mpv.playlist_clear()
+
     def _on_position_changed(self, position):
         self._position = position
         self.position_changed.emit(position)
@@ -178,11 +227,29 @@ class MpvPlayer(AbstractPlayer):
         self.video_format = vformat
 
     def _on_event(self, event):
-        if event['event_id'] == MpvEventID.END_FILE:
+        event_id = event['event_id']
+        if event_id == MpvEventID.END_FILE:
             reason = event['event']['reason']
             logger.debug('Current song finished. reason: %d' % reason)
             if self.state != State.stopped and reason != MpvEventEndFile.ABORTED:
                 self.media_finished.emit()
+        elif event_id == MpvEventID.FILE_LOADED:
+            self.media_loaded.emit()
+        elif event_id == MpvEventID.METADATA_UPDATE:
+            metadata = dict(self._mpv.metadata or {})  # type: ignore
+            logger.debug('metadata updated to %s', metadata)
+            if self._current_metadata.get('__setby__') != 'manual':
+                self._current_metadata['__setby__'] = 'automatic'
+                mapping = Metadata({MetadataFields.title: 'title',
+                                    MetadataFields.album: 'album',
+                                    MetadataFields.artists: 'artist'})
+                for src, tar in mapping.items():
+                    if tar in metadata:
+                        value = metadata[tar]
+                        if src is MetadataFields.artists:
+                            value = [value]
+                        self._current_metadata[src] = value
+                self.metadata_changed.emit(self.current_metadata)
 
     def _set_http_headers(self, http_headers):
         if http_headers:
