@@ -1,15 +1,17 @@
 import asyncio
-import copy
+import warnings
 import logging
 import random
 from enum import IntEnum
 from typing import Optional
 
 from feeluown.excs import ProviderIOError
+from feeluown.utils import aio
 from feeluown.utils.dispatch import Signal
 from feeluown.utils.utils import DedupList
-from feeluown.library.excs import MediaNotFound
-from feeluown.library.model_protocol import SongProtocol
+from feeluown.player import Metadata, MetadataFields
+from feeluown.library import MediaNotFound, SongProtocol, ModelType
+from feeluown.models.uri import reverse
 
 logger = logging.getLogger(__name__)
 
@@ -56,11 +58,6 @@ class Playlist:
         """
         self._app = app
 
-        #: mainthread asyncio loop ref
-        # We know that feeluown is a asyncio-app, and we can assume
-        # that the playlist is inited in main thread.
-        self._loop = asyncio.get_event_loop()
-
         #: init playlist mode normal
         self._mode = PlaylistMode.normal
 
@@ -102,11 +99,15 @@ class Playlist:
         #: When watch mode is on, playlist try to play the mv/video of the song
         self.watch_mode = False
 
+        self._t_scm = self._app.task_mgr.get_or_create('set-current-model')
+
         # .. versionadded:: 3.7.11
         #    The *songs_removed* and *songs_added* signal.
         self.songs_removed = Signal()  # (index, count)
         self.songs_added = Signal()  # (index, count)
 
+        self._app.player.media_finished.connect(self._on_media_finished)
+
     @property
     def mode(self):
         return self._mode
@@ -143,6 +144,29 @@ class Playlist:
         length = len(self._songs)
         self.songs_added.emit(length-1, 1)
 
+    def set_models(self, models, next_=False, fm=False):
+        """
+        .. versionadded: v3.7.13
+        """
+        self.clear()
+        self.batch_add(models)
+        if fm is False:
+            self.mode = PlaylistMode.normal
+        else:
+            self.mode = PlaylistMode.fm
+        if next_ is True:
+            self.next()
+
+    def batch_add(self, models):
+        """
+        .. versionadded: v3.7.13
+        """
+        start_index = len(self._songs)
+        for model in models:
+            self._songs.append(model)
+        end_index = len(self._songs)
+        self.songs_added.emit(start_index, end_index - start_index)
+
     def add(self, song):
         """add song to playlist
 
@@ -205,27 +229,11 @@ class Playlist:
             self._bad_songs.remove(song)
 
     def init_from(self, songs):
-        """
-        THINKING: maybe we should rename this method or maybe we should
-        change mode on application level
-
-        We change playlistmode here because the `player.play_all` call this
-        method. We should check if we need to exit fm mode in `play_xxx`.
-        Currently, we have two play_xxx API: play_all and play_song.
-        1. play_all -> init_from
-        2. play_song -> current_song.setter
-
-        (alpha) temporarily, should only called by player.play_songs
-        """
-
-        if self.mode is PlaylistMode.fm:
-            self.mode = PlaylistMode.normal
-        self.clear()
-        # since we will call songs.clear method during playlist clearing,
-        # we need to deepcopy songs object here.
-        self._songs = DedupList(copy.deepcopy(songs))
-        if self._songs:
-            self.songs_added.emit(0, len(self._songs))
+        warnings.warn(
+            'use set_models instead, this will be removed in v3.8',
+            DeprecationWarning
+        )
+        self.set_models(songs, fm=False)
 
     def clear(self):
         """remove all songs from playlists"""
@@ -262,7 +270,8 @@ class Playlist:
         :param direction: forward if > 0 else backward
         :param loop: regard the song list as a loop
 
-        >>> pl = Playlist([1, 2, 3])
+        >>> from unittest import mock
+        >>> pl = Playlist(mock.Mock(), [1, 2, 3])
         >>> pl._get_good_song()
         1
         >>> pl._get_good_song(base=1)
@@ -330,7 +339,7 @@ class Playlist:
             return self._get_good_song(base=-1, direction=-1)
 
         if self.playback_mode == PlaybackMode.random:
-            previous_song = self._get_good_song(direction=-1)
+            previous_song = self._get_good_song(direction=-1, random_=True)
         else:
             current_index = self._songs.index(self.current_song)
             previous_song = self._get_good_song(base=current_index - 1, direction=-1)
@@ -343,6 +352,13 @@ class Playlist:
         else:
             return self.set_current_song(self.next_song)
 
+    def _on_media_finished(self):
+        # Play next model when current media is finished.
+        if self.playback_mode == PlaybackMode.one_loop:
+            return self.set_current_song(self.current_song)
+        else:
+            self.next()
+
     def previous(self) -> Optional[asyncio.Task]:
         """return to the previous song in playlist"""
         return self.set_current_song(self.previous_song)
@@ -380,8 +396,7 @@ class Playlist:
 
         # FIXME(cosven): `current_song.setter` depends on app.task_mgr and app.library,
         # which make it hard to test.
-        task_spec = self._app.task_mgr.get_or_create('set-current-song')
-        return task_spec.bind_coro(self.a_set_current_song(song))
+        return self._t_scm.bind_coro(self.a_set_current_song(song))
 
     async def a_set_current_song(self, song):
         song_str = f'{song.source}:{song.title_display} - {song.artists_name_display}'
@@ -444,6 +459,27 @@ class Playlist:
         self.song_changed.emit(song)
         self.song_changed_v2.emit(song, media)
 
+        if song is not None:
+            if media is None:
+                self.next()
+            else:
+                # Note that the value of model v1 {}_display may be None.
+                metadata = Metadata({
+                    MetadataFields.uri: reverse(song),
+                    MetadataFields.source: song.source,
+                    MetadataFields.title: song.title_display or '',
+                    # The song.artists_name should return a list of strings
+                    MetadataFields.artists: [song.artists_name_display or ''],
+                    MetadataFields.album: song.album_name_display or '',
+                })
+                kwargs = {}
+                if not self._app.has_gui:
+                    kwargs['video'] = False
+                # TODO: set artwork field
+                self._app.player.play(media, metadata=metadata, **kwargs)
+        else:
+            self._app.player.stop()
+
     async def _prepare_media(self, song):
         task_spec = self._app.task_mgr.get_or_create('prepare-media')
         if self.watch_mode is True:
@@ -462,3 +498,66 @@ class Playlist:
                 return mv_media
         return await task_spec.bind_blocking_io(
             self._app.library.song_prepare_media, song, self.audio_select_policy)
+
+    def set_current_model(self, model):
+        """
+        .. versionadded: 3.7.13
+        """
+        if ModelType(model.meta.model_type) is ModelType.song:
+            return self.set_current_song(model)
+        if model is None:
+            self._app.player.stop()
+            return
+        return self._t_scm.bind_coro(self.a_set_current_model(model))
+
+    async def a_set_current_model(self, model):
+        """
+        TODO: handle when model is a song
+
+        .. versionadded: 3.7.13
+        """
+        assert ModelType(model.meta.model_type) is ModelType.video, \
+            "{model.meta.model_type} is not supported, expecting a video model, "
+
+        video = model
+        try:
+            media = await aio.run_fn(
+                self._app.library.video_prepare_media,
+                video,
+                self._app.config.VIDEO_SELECT_POLICY
+            )
+        except MediaNotFound:
+            self._app.show_msg('没有可用的播放链接')
+        else:
+            metadata = Metadata({
+                # The value of model v1 title_display may be None.
+                MetadataFields.title: video.title_display or '',
+                MetadataFields.source: video.source,
+                MetadataFields.uri: reverse(video),
+            })
+            kwargs = {}
+            if not self._app.has_gui:
+                kwargs['video'] = False
+            self._app.player.play(media, metadata=metadata, **kwargs)
+
+    """
+    Sync methods.
+
+    Currently, playlist has both async and sync methods to keep backward
+    compatibility. Sync methods will be replaced by async methods in the end.
+    """
+    def play_model(self, model):
+        """Set current model and play it
+
+        .. versionadded: 3.7.14
+        """
+        task = self.set_current_model(model)
+        if task is not None:
+            def cb(future):
+                try:
+                    future.result()
+                except:  # noqa
+                    logger.exception('play model failed')
+                else:
+                    self._app.player.resume()
+            task.add_done_callback(cb)
